import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from . import constants as con
from . import helper_functions as hfunc


def plot_ascent(time, height, speed, v_y, a_y, unit="m"):
    """
    Plot ascent data including height, speed, and vertical acceleration.

    Args:
    - time (pd.Series): Time series data.
    - height (pd.Series): Height series data.
    - speed (pd.Series): Speed series data.
    - v_y (pd.Series): Vertical velocity series data.
    - a_y (pd.Series): Vertical acceleration series data.
    - unit (str): Unit of measurement for height, speed, etc.
    """
    fig, ax1 = plt.subplots()
    ax1.plot(time, height, color="b")
    ax1.set_xlabel("Time (s)")
    ax1.set_ylabel(f"Height ({unit})", color="b")
    ax1.tick_params(axis="y", labelcolor="b")

    ax2 = ax1.twinx()
    ax2.plot(time, v_y, color="r")
    ax2.set_ylabel(f"Vertical velocity ({unit}/s)", color="r")
    ax2.tick_params(axis="y", labelcolor="r")

    ax3 = ax1.twinx()
    ax3.spines["right"].set_position(("outward", 60))
    ax3.plot(time, a_y, color="g")
    ax3.set_ylabel(f"Vertical acceleration ({unit}/s^2)", color="g")
    ax3.tick_params(axis="y", labelcolor="g")

    plt.title("Ascent Overview")
    plt.show()


def plot_aerodynamics(time, height, speed, q, angle_to_vertical, air_density, unit="m"):
    """
    Plot aerodynamic parameters over the ascent.

    Args:
    - time (pd.Series): Time series data.
    - height (pd.Series): Height series data.
    - speed (pd.Series): Speed series data.
    - q (pd.Series): Dynamic pressure series data.
    - angle_to_vertical (pd.Series): Angle to vertical series data.
    - air_density (pd.Series): Air density series data.
    - unit (str): Unit of measurement for height, speed, etc.
    """
    fig, ax1 = plt.subplots()
    ax1.plot(time, height, color="b", label="Height")
    ax1.set_xlabel("Time (s)")
    ax1.set_ylabel(f"Height ({unit})", color="b")
    ax1.tick_params(axis="y", labelcolor="b")
    ax1.set_ylim(0, height.max() * 1.1)

    ax2 = ax1.twinx()
    ax2.plot(time, speed, color="r", label="Speed")
    ax2.set_ylabel(f"Speed ({unit}/s)", color="r")
    ax2.tick_params(axis="y", labelcolor="r")
    ax2.set_ylim(0, speed.max() * 1.1)

    ax3 = ax1.twinx()
    ax3.spines["right"].set_position(("outward", 60))
    ax3.plot(time, q / 1000, color="g", label="q")
    ax3.set_ylabel("q (kPa)", color="g")
    ax3.tick_params(axis="y", labelcolor="g")
    ax3.set_ylim(0, q.max() / 1000 * 1.1)

    ax4 = ax1.twinx()
    ax4.spines["right"].set_position(("outward", 120))
    ax4.plot(time, angle_to_vertical * 180 / np.pi, color="c", label="Angle to Vertical")
    ax4.set_ylabel(f"Angle to vertical (deg)", color="c")
    ax4.tick_params(axis="y", labelcolor="c")
    ax4.set_yticks(range(0, 91, 15))

    ax5 = ax1.twinx()
    ax5.spines["right"].set_position(("outward", 180))
    ax5.plot(time, air_density, color="m", label="Air Density")
    ax5.set_ylabel("Air Density (kg/m^3)", color="m")
    ax5.tick_params(axis="y", labelcolor="m")

    plt.title("Aerodynamics during Ascent")
    plt.show()


def plot_airbrakes_ascent(ascent, airbrakes, unit="m"):
    """
    Plot ascent data including height, speed, acceleration, deployment angle, and force on airbrakes.

    Args:
    - ascent (pd.DataFrame): Dataframe containing the ascent data with airbrakes.
    - unit (str): Unit of measurement for height, speed, etc.
    """

    # Existing code for height, speed, and acceleration plots
    height = (
        ascent["height"].copy() * con.m_to_ft_conversion
        if unit == "ft"
        else ascent["height"].copy()
    )
    speed = (
        ascent["speed"].copy() * con.m_to_ft_conversion
        if unit == "ft"
        else ascent["speed"].copy()
    )
    accel = (
        ascent["a_y"].copy() * con.m_to_ft_conversion
        if unit == "ft"
        else ascent["a_y"].copy()
    )

    # New code to calculate force
    # Assuming 'ascent' DataFrame includes 'q' and 'deployment_angle'
    # and 'airbrake_deployment' is defined or calculated before this function
    force = (ascent["q"]) * airbrakes.A_flap * np.sin(ascent["deployment_angle"]) * 3

    fig, ax1 = plt.subplots()

    # Existing plotting code for height, speed, acceleration, and deployment angle
    ax1.plot(ascent["time"], height, color="b")
    ax1.set_xlabel("Time (s)")
    ax1.set_ylabel(f"Height ({unit})", color="b")
    ax1.tick_params(axis="y", labelcolor="b")
    ax1.axhline(y=10000, color='gray', linestyle='--')

    ax2 = ax1.twinx()
    ax2.plot(ascent["time"], speed, color="r")
    ax2.set_ylabel(f"Speed ({unit}/s)", color="r")
    ax2.tick_params(axis="y", labelcolor="r")

    ax3 = ax1.twinx()
    ax3.spines["right"].set_position(("outward", 60))
    ax3.plot(ascent["time"], accel, color="g")
    ax3.set_ylabel(f"Acceleration ({unit}/s^2)", color="g")
    ax3.tick_params(axis="y", labelcolor="g")

    ax4 = ax1.twinx()
    ax4.spines["right"].set_position(("outward", 120))
    ax4.plot(ascent["time"], np.rad2deg(ascent["deployment_angle"]), color="y")
    ax4.set_ylabel(f"Deployment Angle (deg)", color="y")
    ax4.tick_params(axis="y", labelcolor="y")
    ax4.set_yticks(range(0, 46, 15))

    # New plotting code for force
    ax5 = ax1.twinx()
    ax5.spines["right"].set_position(("outward", 180))  # Adjust this as needed
    ax5.plot(ascent["time"], force, color="m")  # 'm' for magenta, can choose any color
    ax5.set_ylabel("Force (N)", color="m")  # Assuming force is in Newtons
    ax5.tick_params(axis="y", labelcolor="m")

    plt.title("Ascent with Airbrakes")
    plt.show()


def display_apogee_parameters_table(ascent, parameters_at_flight_events, unit="m"):
    """
    Display a table of parameters at apogee, including comparisons between ascent with and without airbrakes.

    Args:
    - ascent (pd.DataFrame): Dataframe containing the ascent data with airbrakes.
    - parameters_at_flight_events (pd.DataFrame): Dataframe containing parameters at key flight events.
    - unit (str): Unit of length.
    """
    # Calculate the relevant parameters for ascent with airbrakes
    last_index = len(ascent) - 1
    time_with_airbrakes = ascent["time"].iloc[last_index]
    height_with_airbrakes = (
        ascent["height"].iloc[last_index] * con.m_to_ft_conversion
        if unit == "ft"
        else ascent["height"].iloc[last_index]
    )
    speed_with_airbrakes = (
        ascent["speed"].iloc[last_index] * con.m_to_ft_conversion
        if unit == "ft"
        else ascent["speed"].iloc[last_index]
    )
    accel_with_airbrakes = (
        ascent["a_y"].iloc[last_index] * con.m_to_ft_conversion
        if unit == "ft"
        else ascent["a_y"].iloc[last_index]
    )
    g_force_with_airbrakes = abs(ascent["a_y"].iloc[last_index] / con.F_gravity)

    # Extract parameters at apogee without airbrakes
    time_without_airbrakes = parameters_at_flight_events.loc["Apogee", "Time (s)"]
    height_without_airbrakes = parameters_at_flight_events.loc[
        "Apogee", f"Height ({unit})"
    ]
    speed_without_airbrakes = parameters_at_flight_events.loc[
        "Apogee", f"Speed ({unit}/s)"
    ]
    accel_without_airbrakes = parameters_at_flight_events.loc[
        "Apogee", f"Vertical accel ({unit}/s^2)"
    ]
    g_force_without_airbrakes = parameters_at_flight_events.loc[
        "Apogee", f"G-force (g)"
    ]

    # Create DataFrame for comparison
    parameters_at_apogee = pd.DataFrame(
        {
            "Time (s)": [time_with_airbrakes, time_without_airbrakes],
            f"Height ({unit})": [height_with_airbrakes, height_without_airbrakes],
            f"Speed ({unit}/s)": [speed_with_airbrakes, speed_without_airbrakes],
            f"Accel ({unit}/s^2)": [accel_with_airbrakes, accel_without_airbrakes],
            f"G-force (g)": [g_force_with_airbrakes, g_force_without_airbrakes],
        },
        index=["Apogee with Airbrakes", "Apogee without Airbrakes"],
    )

    # Calculate deltas and percentage deltas
    delta_values = (
        parameters_at_apogee.loc["Apogee with Airbrakes"]
        - parameters_at_apogee.loc["Apogee without Airbrakes"]
    )
    parameters_at_apogee.loc["Delta"] = delta_values
    percent_delta_values = round(
        delta_values / parameters_at_apogee.loc["Apogee without Airbrakes"] * 100, 2
    )
    parameters_at_apogee.loc["% Delta"] = percent_delta_values

    return parameters_at_apogee


def create_flight_event_table(
    time,
    height,
    v_y,
    speed,
    a_y,
    g_force,
    Ma,
    dataset,
    liftoff_index,
    launch_rail_cleared_index,
    burnout_index,
    apogee_index,
    unit="m",
):
    """
    Create a table of parameters at key flight events.

    Args:
    - time, height, v_y, speed, a_y, g_force, Ma (pd.Series): Data series.
    - dataset (pd.DataFrame): The dataset containing additional flight data.
    - liftoff_index, launch_rail_cleared_index, burnout_index, apogee_index (int): Indices of key flight events.
    - unit (str): Unit of measurement for height, speed, etc.

    Returns:
    - pd.DataFrame: Table of parameters at key flight events.
    """
    max_g_index = g_force.idxmax()
    max_speed_Ma_index = speed.idxmax()
    max_q_index = dataset["q"][:apogee_index].idxmax()

    parameters_at_flight_events = pd.DataFrame(
        {
            "Time (s)": [
                round(time.iloc[liftoff_index], 2),
                round(time.iloc[launch_rail_cleared_index], 2),
                round(time.iloc[max_g_index], 2),
                round(time.iloc[max_q_index], 2),
                round(time.iloc[max_speed_Ma_index], 2),
                round(time.iloc[burnout_index], 2),
                round(time.iloc[-1], 2),
            ],
            f"Height ({unit})": [
                round(height.iloc[liftoff_index], 2),
                round(height.iloc[launch_rail_cleared_index], 2),
                round(height.iloc[max_g_index], 2),
                round(height.iloc[max_q_index], 2),
                round(height.iloc[max_speed_Ma_index], 2),
                round(height.iloc[burnout_index], 2),
                round(height.iloc[-1], 2),
            ],
            f"Vertical velocity ({unit}/s)": [
                round(v_y.iloc[liftoff_index], 2),
                round(v_y.iloc[launch_rail_cleared_index], 2),
                round(v_y.iloc[max_g_index], 2),
                round(v_y.iloc[max_q_index], 2),
                round(v_y.iloc[max_speed_Ma_index], 2),
                round(v_y.iloc[burnout_index], 2),
                round(v_y.iloc[-1], 2),
            ],
            f"Speed ({unit}/s)": [
                round(speed.iloc[liftoff_index], 2),
                round(speed.iloc[launch_rail_cleared_index], 2),
                round(speed.iloc[max_g_index], 2),
                round(speed.iloc[max_q_index], 2),
                round(speed.iloc[max_speed_Ma_index], 2),
                round(speed.iloc[burnout_index], 2),
                round(speed.iloc[-1], 2),
            ],
            f"Vertical accel ({unit}/s^2)": [
                round(a_y.iloc[liftoff_index], 2),
                round(a_y.iloc[launch_rail_cleared_index], 2),
                round(a_y.iloc[max_g_index], 2),
                round(a_y.iloc[max_q_index], 2),
                round(a_y.iloc[max_speed_Ma_index], 2),
                round(a_y.iloc[burnout_index], 2),
                round(a_y.iloc[-1], 2),
            ],
            f"G-force (g)": [
                round(g_force.iloc[liftoff_index], 2),
                round(g_force.iloc[launch_rail_cleared_index], 2),
                round(g_force.iloc[max_g_index], 2),
                round(g_force.iloc[max_q_index], 2),
                round(g_force.iloc[max_speed_Ma_index], 2),
                round(g_force.iloc[burnout_index], 2),
                round(g_force.iloc[-1], 2),
            ],
            f"Mach": [
                round(
                    hfunc.mach_number_fn(
                        dataset["speed"].iloc[liftoff_index],
                        dataset["temperature"].iloc[liftoff_index],
                    ),
                    3,
                ),
                round(
                    hfunc.mach_number_fn(
                        dataset["speed"].iloc[launch_rail_cleared_index],
                        dataset["temperature"].iloc[launch_rail_cleared_index],
                    ),
                    3,
                ),
                round(
                    hfunc.mach_number_fn(
                        dataset["speed"].iloc[max_g_index],
                        dataset["temperature"].iloc[max_g_index],
                    ),
                    3,
                ),
                round(
                    hfunc.mach_number_fn(
                        dataset["speed"].iloc[max_q_index],
                        dataset["temperature"].iloc[max_q_index],
                    ),
                    3,
                ),
                round(
                    hfunc.mach_number_fn(
                        dataset["speed"].iloc[max_speed_Ma_index],
                        dataset["temperature"].iloc[max_speed_Ma_index],
                    ),
                    3,
                ),
                round(
                    hfunc.mach_number_fn(
                        dataset["speed"].iloc[burnout_index],
                        dataset["temperature"].iloc[burnout_index],
                    ),
                    3,
                ),
                round(
                    hfunc.mach_number_fn(
                        dataset["speed"].iloc[-1], dataset["temperature"].iloc[-1]
                    ),
                    3,
                ),
            ],
            f"q (kPa)": [
                round(dataset["q"].iloc[liftoff_index] / 1000, 2),
                round(dataset["q"].iloc[launch_rail_cleared_index] / 1000, 2),
                round(dataset["q"].iloc[max_g_index] / 1000, 2),
                round(dataset["q"].iloc[max_q_index] / 1000, 2),
                round(dataset["q"].iloc[max_speed_Ma_index] / 1000, 2),
                round(dataset["q"].iloc[burnout_index] / 1000, 2),
                round(dataset["q"].iloc[-1] / 1000, 2),
            ],
        },
        index=[
            "Liftoff",
            "Off Launch Rail",
            "Max g-Force",
            "Max q",
            "Max Speed, Ma",
            "Burnout",
            "Apogee",
        ],
    ).sort_values(by=["Time (s)"])

    return parameters_at_flight_events



# Import libraries, define natural constants, helper functions
import sys
import os

import pandas as pd
import numpy as np

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'examples')))
from . import helper_functions as hfunc
from . import constants as con

default_timestep = 0.02
""" Notes on timesteps:

The default for OpenRocket sims is 0.01s for the first while, and then somewhere between 0.02 and 0.05 for a while, and then 0.05 for most of the rest of the ascent. It simulates more complicated dynamics than we do

A timestep of 0.02s gives apogees a few feet different for a 10k launch compared to using 0.001s. 0.001s can still be used for one-off sims, but when running many sims, 0.02s is better.
"""
# Flight simulation function
def simulate_flight(rocket, launch_conditions, timestep=default_timestep):
    """
    Simulate the flight of a rocket until its apogee given its specifications and launch conditions.

    Args:
    - rocket (Rocket): An instance of the Rocket class.
    - launch_conditions (LaunchConditions): An instance of the LaunchConditions class.
    - timestep (float): The time increment for the simulation in seconds.

    Returns:
    - tuple: A tuple containing the dataset of simulation results and indices of key flight events.
    """

    # Initialize rocket parameters and launch conditions
    dry_mass = rocket.dry_mass
    fuel_mass_lookup = rocket.motor.fuel_mass_curve
    engine_thrust_lookup = rocket.motor.thrust_curve
    Cd_A_rocket_fn = rocket.Cd_A_rocket
    burnout_time = rocket.motor.burn_time

    # Extract launch condition parameters
    launchpad_pressure = launch_conditions.launchpad_pressure
    launchpad_temp = launch_conditions.launchpad_temp
    L_launch_rail = launch_conditions.L_launch_rail
    launch_angle = launch_conditions.launch_angle
    
    T_lapse_rate = launch_conditions.local_T_lapse_rate
    F_gravity = launch_conditions.local_gravity

    # Initialize simulation variables
    time, height, speed, a_y, a_x, v_y, v_x, Cd_A_rocket, Ma, q = 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    angle_to_vertical = np.deg2rad(90 - launch_angle)
    # Calculate constants to be used in air density function, set initial air density
    multiplier = launchpad_pressure / (con.R_specific_air * pow(launchpad_temp, - F_gravity / (con.R_specific_air * T_lapse_rate)))
    exponent = - F_gravity / (con.R_specific_air * T_lapse_rate) - 1
    air_density = hfunc.air_density_optimized(launchpad_temp, multiplier, exponent)

    # Store the initial state of the rocket
    simulated_values = [
        [
            time,
            height,
            speed,
            a_y,
            a_x,
            v_y,
            v_x,
            launchpad_temp,
            air_density,
            q,
            Ma,
            Cd_A_rocket, # why is this being saved?
            angle_to_vertical,
        ]
    ]

    # Simulate motor burn until liftoff
    while (
        hfunc.thrust_at_time(time, engine_thrust_lookup)
        / hfunc.mass_at_time(time, dry_mass, fuel_mass_lookup)
        <= F_gravity
    ):
        time += timestep
        # Append simulation values for each timestep
        simulated_values.append(
            [
                time,
                height, # 0
                speed, # 0
                a_y, # 0
                a_x, # 0
                v_y, # 0
                v_x, # 0
                launchpad_temp,
                air_density,
                q, # 0
                Ma, # 0
                Cd_A_rocket, # 0
                angle_to_vertical,
            ]
        )

    liftoff_index = len(simulated_values)

    # Liftoff until launch rail cleared
    time += timestep
    effective_L_launch_rail = L_launch_rail - rocket.h_second_rail_button
    effective_h_launch_rail = effective_L_launch_rail * np.cos(angle_to_vertical)

    # Simulate flight from liftoff until the launch rail is cleared
    while height < effective_h_launch_rail:
        # Update environmental conditions based on height
        temperature = hfunc.temp_at_height(height, launchpad_temp, lapse_rate = T_lapse_rate)
        air_density = hfunc.air_density_optimized(temperature, multiplier,exponent)

        # Calculate Mach number, drag coefficient, and forces
        Ma = hfunc.mach_number_fn(speed, temperature)
        Cd_A_rocket = Cd_A_rocket_fn(Ma)
        q = hfunc.calculate_dynamic_pressure(air_density, speed)
        F_drag = q * Cd_A_rocket

        # Update rocket's motion parameters
        mass = hfunc.mass_at_time(time, dry_mass, fuel_mass_lookup)
        thrust = hfunc.thrust_at_time(time, engine_thrust_lookup)
        a_y = (thrust - F_drag) * np.cos(angle_to_vertical) / mass - F_gravity
        a_x = (thrust - F_drag) * np.sin(angle_to_vertical) / mass
        v_y += a_y * timestep
        v_x += a_x * timestep
        speed = np.sqrt(v_y**2 + v_x**2)
        height += v_y * timestep

        # Append updated simulation values
        simulated_values.append(
            [
                time,
                height,
                speed,
                a_y,
                a_x,
                v_y,
                v_x,
                temperature,
                air_density,
                q,
                Ma,
                Cd_A_rocket,
                angle_to_vertical,
            ]
        )
        time += timestep

    launch_rail_cleared_index = len(simulated_values)

    # Flight from launch rail cleared until burnout
    while time < burnout_time:
        # Update environmental conditions based on height
        temperature = hfunc.temp_at_height(height, launchpad_temp, lapse_rate = T_lapse_rate)
        air_density = hfunc.air_density_optimized(temperature, multiplier,exponent)

        # Calculate Mach number, Drag coefficient, and Forces
        Ma = hfunc.mach_number_fn(speed, temperature)
        Cd_A_rocket = Cd_A_rocket_fn(Ma)
        q = hfunc.calculate_dynamic_pressure(air_density, speed)
        F_drag = q * Cd_A_rocket

        # Update rocket's motion parameters
        mass = hfunc.mass_at_time(time, dry_mass, fuel_mass_lookup)
        thrust = hfunc.thrust_at_time(time, engine_thrust_lookup)
        a_y = (thrust - F_drag) * np.cos(angle_to_vertical) / mass - F_gravity
        a_x = (thrust - F_drag) * np.sin(angle_to_vertical) / mass
        v_y += a_y * timestep
        v_x += a_x * timestep
        speed = np.sqrt(v_y**2 + v_x**2)
        height += v_y * timestep

        # Recalculate the angle to the veritcal
        angle_to_vertical = np.arctan(v_x / v_y)

        # Append updated simulation values
        simulated_values.append(
            [
                time,
                height,
                speed,
                a_y,
                a_x,
                v_y,
                v_x,
                temperature,
                air_density,
                q,
                Ma,
                Cd_A_rocket,
                angle_to_vertical,
            ]
        )

        time += timestep

    burnout_index = len(simulated_values)

    # Flight from burnout to apogee
    mass = dry_mass

    while v_y > 0:
        temperature = hfunc.temp_at_height(height, launchpad_temp, lapse_rate = T_lapse_rate)
        air_density = hfunc.air_density_optimized(temperature, multiplier,exponent)

        # Calculate Mach number, Drag coefficient, and Forces
        Ma = hfunc.mach_number_fn(speed, temperature)
        Cd_A_rocket = Cd_A_rocket_fn(Ma)
        q = hfunc.calculate_dynamic_pressure(air_density, speed)
        F_drag = q * Cd_A_rocket

        # Update rocket's motion parameters
        a_y = -F_drag * np.cos(angle_to_vertical) / mass - F_gravity
        a_x = -F_drag * np.sin(angle_to_vertical) / mass
        v_y += a_y * timestep
        v_x += a_x * timestep
        speed = np.sqrt(v_y**2 + v_x**2)
        height += v_y * timestep

        # Recalculate the angle to the veritcal
        angle_to_vertical = np.arctan(v_x / v_y)

        # Append updated simulation values
        simulated_values.append(
            [
                time,
                height,
                speed,
                a_y,
                a_x,
                v_y,
                v_x,
                temperature,
                air_density,
                q,
                Ma,
                Cd_A_rocket,
                angle_to_vertical,
            ]
        )

        time += timestep

    # Mark the index at apogee (highest point)
    apogee_index = len(simulated_values)
    simulated_values[-1][12] = simulated_values[-2][12]

    # Convert the list of simulation values to a DataFrame for easier analysis and visualization
    dataset = pd.DataFrame(
        {
            "time": [row[0] for row in simulated_values],
            "height": [row[1] for row in simulated_values],
            "speed": [row[2] for row in simulated_values],
            "a_y": [row[3] for row in simulated_values],
            "a_x": [row[4] for row in simulated_values],
            "v_y": [row[5] for row in simulated_values],
            "v_x": [row[6] for row in simulated_values],
            "temperature": [row[7] for row in simulated_values],
            "air_density": [row[8] for row in simulated_values],
            "q": [row[9] for row in simulated_values],
            "Ma": [row[10] for row in simulated_values],
            "Cd_A_rocket": [row[11] for row in simulated_values],
            "angle_to_vertical": [row[12] for row in simulated_values],
        }
    )

    return (
        dataset,
        liftoff_index,
        launch_rail_cleared_index,
        burnout_index,
        apogee_index
    )


# Flight with airbrakes simulation function
def simulate_airbrakes_flight(pre_brake_flight, rocket, launch_conditions, airbrakes, timestep = default_timestep):
    """
    Simulate the flight of a rocket during its post-burnout ascent with airbrakes deployed until apogee, given its specifications and environmental conditions.

    Args:
    - pre_brake_flight (DataFrame): A DataFrame containing the simulation results from the rocket's ascent until burnout. # TODO: change to a tuple of the dataset at the time to start the simulation
    - rocket (Rocket): An instance of the Rocket class.
    - launch_conditions (LaunchConditions): An instance of the LaunchConditions class.
    - airbrakes (Airbrakes): An instance of the Airbrakes class.
    - timestep (float): The time increment for the simulation in seconds.

    Returns:
    - tuple: A tuple containing the dataset of simulation results and the time at which the airbrakes are fully deployed.
    """
    # Extract rocket parameters
    mass = rocket.dry_mass
    Cd_A_rocket_fn = rocket.Cd_A_rocket

    # Extract launch condition parameters
    launchpad_temp = pre_brake_flight.temperature.iloc[0]
    launchpad_pressure = pre_brake_flight.air_density.iloc[0] * con.R_specific_air * launchpad_temp
    
    T_lapse_rate = launch_conditions.local_T_lapse_rate    
    F_gravity = launch_conditions.local_gravity

    # Calculate constants to be used in air density function
    multiplier = launchpad_pressure / (con.R_specific_air * pow(launchpad_temp, - F_gravity / (con.R_specific_air * T_lapse_rate)))
    exponent = - F_gravity / (con.R_specific_air * T_lapse_rate) - 1

    # Extract airbrakes parameters
    Cd_brakes = airbrakes.Cd_brakes
    max_deployment_speed = np.deg2rad(airbrakes.max_deployment_speed)
    max_deployment_angle = np.deg2rad(airbrakes.max_deployment_angle)
    A_brakes = airbrakes.num_flaps * airbrakes.A_flap

    pre_brake_flight["deployment_angle"] = 0

    # Initialize simulation variables
    height = pre_brake_flight["height"].iloc[-1]
    speed = pre_brake_flight["speed"].iloc[-1]
    v_y = pre_brake_flight["v_y"].iloc[-1]
    v_x = pre_brake_flight["v_x"].iloc[-1]
    time = pre_brake_flight["time"].iloc[-1]
    angle_to_vertical = np.arctan(v_x / v_y)

    deployment_angle = 0
    time_recorded = False

    # for efficiency, may be removed if/when the simulation is made more accurate by the cd of the brakes changing during the sim:
    A_Cd_brakes = A_brakes * Cd_brakes

    simulated_values = []
    while v_y > 0:
        time += timestep

        temperature = hfunc.temp_at_height(height, launchpad_temp, lapse_rate = T_lapse_rate)
        air_density = hfunc.air_density_optimized(temperature, multiplier,exponent)
        Ma = hfunc.mach_number_fn(speed, temperature)
        Cd_A_rocket = Cd_A_rocket_fn(Ma)
        q = hfunc.calculate_dynamic_pressure(air_density, speed)

        deployment_angle = min(max_deployment_angle, deployment_angle + max_deployment_speed * timestep)

        if deployment_angle >= max_deployment_angle and not time_recorded:
            time_of_max_deployment = time
            time_recorded = True


        F_drag = q * (np.sin(deployment_angle) * A_Cd_brakes + Cd_A_rocket)
        a_y = -F_drag * np.cos(angle_to_vertical) / mass - F_gravity
        a_x = -F_drag * np.sin(angle_to_vertical) / mass
        v_y += a_y * timestep
        v_x += a_x * timestep
        speed = np.sqrt(v_y**2 + v_x**2)
        height += v_y * timestep

        angle_to_vertical = np.arctan(v_x / v_y)

        simulated_values.append(
            [
                time,
                height,
                speed,
                a_y,
                a_x,
                v_y,
                v_x,
                temperature,
                air_density,
                q,
                Ma,
                Cd_A_rocket,
                angle_to_vertical,
                deployment_angle,
            ]
        )

    simulated_values[-1][12] = simulated_values[-2][12]

    data = {
        "time": [row[0] for row in simulated_values],
        "height": [row[1] for row in simulated_values],
        "speed": [row[2] for row in simulated_values],
        "a_y": [row[3] for row in simulated_values],
        "a_x": [row[4] for row in simulated_values],
        "v_y": [row[5] for row in simulated_values],
        "v_x": [row[6] for row in simulated_values],
        "temperature": [row[7] for row in simulated_values],
        "air_density": [row[8] for row in simulated_values],
        "q": [row[9] for row in simulated_values],
        "Ma": [row[10] for row in simulated_values],
        "Cd_A_rocket": [row[11] for row in simulated_values],
        "angle_to_vertical": [row[12] for row in simulated_values],
        "deployment_angle": [row[13] for row in simulated_values],
    }

    ascent = pd.concat([pre_brake_flight, pd.DataFrame(data)], ignore_index=True)

    if not time_recorded:
        time_of_max_deployment = time

    return ascent, time_of_max_deployment


# TODO: move this to a test file
    # run a couple hundred different timesteps in logspace between 0.001 and 0.1 to see how it changes to help pick a good timestep
    
"""    apogees = []
    for timestep in np.logspace(-3, -1, 200):
        dataset, liftoff_index, launch_rail_cleared_index, burnout_index, apogee_index = simulate_flight(rocket=Juno3_rocket, timestep=timestep)
        ascent, time_of_max_deployment = simulate_airbrakes_flight(dataset.iloc[:burnout_index].copy(), rocket=Juno3_rocket, launch_conditions=Juno3_launch_conditions, timestep=0.001)
        apogees.append(ascent["height"].iloc[-1]*3.28084)
        print(len(apogees))
    # plot them
    import matplotlib.pyplot as plt
    plt.plot(np.logspace(-3, -1, 200), apogees)
    plt.xscale("log")
    plt.xlabel("Timestep (s)")
    plt.ylabel("Apogee (ft)")
    plt.title("Apogee vs Timestep")
    plt.show()"""
    


import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import unittest
# from rocketflightsim import constants as con
# from rocketflightsim import helper_functions as hfunc
from rocketflightsim.rocket_classes import Motor, Rocket, LaunchConditions
from rocketflightsim.flight_simulation import simulate_flight, simulate_airbrakes_flight

from test_configs import past_flights

class TestFlightSimulation(unittest.TestCase):
    def test_flight_simulation(self):
            print("Testing flight simulation")

            for flight in past_flights:
                dataset, liftoff_index, launch_rail_cleared_index, burnout_index, apogee_index = simulate_flight(rocket=flight.rocket, launch_conditions=flight.launch_conditions, timestep=0.001)

                apogee_simulated = dataset["height"].iloc[apogee_index - 1]
                apogee_actual = flight.apogee
                difference = abs(apogee_simulated - apogee_actual)
                proportional_difference = difference / apogee_actual
                print(f"Flight {flight.name}: \n\tSimulated apogee: {round(apogee_simulated,2)} m\n\tActual apogee: {round(apogee_actual,2)} m\n\tDifference: {round(difference,2)} m\n\tPercent difference: {round(proportional_difference*100,2)}%\n")

                assert proportional_difference < 0.1
            print("\n--------------------")

class TestAirbrakesFlightSimulation(unittest.TestCase):
     def test_airbrakes_flight_simulation(self):
            print("\n--------------------")
            print("Testing airbrakes flight simulation")
            from test_configs import Juno3_rocket, Juno3_launch_conditions
            from example_configurations import default_airbrakes_model
    
            dataset, liftoff_index, launch_rail_cleared_index, burnout_index, apogee_index = simulate_flight(rocket=Juno3_rocket, launch_conditions=Juno3_launch_conditions, timestep=0.001)
            print(f"Burnout: \n\tHeight: {dataset['height'].iloc[burnout_index - 1]} m\n\tSpeed: {dataset['speed'].iloc[burnout_index - 1]} m/s\n\tTime: {dataset['time'].iloc[burnout_index - 1]} s\n")
            print(f"No-airbraking apogee: \n\tHeight: {dataset['height'].iloc[apogee_index - 1]} m\n\tTime: {dataset['time'].iloc[apogee_index - 1]} s\n")
            # TODO: make it take the actual flight data at burnout and sim from there
            ascent, time_of_max_deployment = simulate_airbrakes_flight(dataset.iloc[:burnout_index].copy(), rocket=Juno3_rocket, launch_conditions=Juno3_launch_conditions, airbrakes=default_airbrakes_model, timestep=0.001)
            print(f"Airbrakes apogee: \n\tHeight: {ascent['height'].iloc[-1]} m\n\tTime: {ascent['time'].iloc[-1]} s\n\n")
            print("--------------------")

            assert ascent["height"].iloc[-1] > 1000
            assert ascent["height"].iloc[-1] < dataset["height"].iloc[apogee_index - 1]
            assert ascent["time"].iloc[-1] < dataset["time"].iloc[apogee_index - 1]
            # assert ascent["speed"].iloc[-1] < dataset["speed"].iloc[apogee_index - 1] ?



if __name__ == '__main__':
    unittest.main()


import numpy as np
from . import constants as con

# aerodynamics
def temp_at_height(h, launchpad_temp, lapse_rate = con.T_lapse_rate):
    """
    Calculate the temperature at a given height above the launchpad. Within the troposphere, the temperature decreases linearly with height at a rate known as the lapse rate. The lapse rate is typically around 6.5 degrees Celsius per kilometer. 

    Args:
    - h (float): Height above the launchpad in meters.
    - launchpad_temp (float): Temperature at the launchpad in Celsius or Kelvin.
    - lapse_rate (float, optional): Rate at which temperature decreases with height in degrees Celsius or Kelvin per meter. Defaults to the standard lapse rate of 0.0065.

    Returns:
    - float: Temperature at the given height in Celsius or Kelvin (same as input).
    """
    return launchpad_temp + (h * lapse_rate)

def air_density_optimized(temp, multiplier, exponent):
    """
    Calculate the density of air at a given height above the launchpad.

    Args:
    - temp (float): Temperature at the given height in Kelvin.
    - multiplier (float): A constant derived from the temperature and pressure at the launchpad, the lapse rate, the specific gas constant for air, and the magnitude of the force of gravity. Calculated at the start of the simulation script outside of the main loops. Equal to P_launchpad / (R_air * pow(T_launchpad, - F_gravity / (R_air * T_lapse_rate))).
    - exponent (float): A constant derived from the lapse rate, the specific gas constant for air, and the magnitude of the force of gravity. Calculated at the start of the simulation script outside of the main loops. Equal to - F_gravity / (R_air * T_lapse_rate) - 1.

    Returns:
    - float: Air density at the given height in kilograms per cubic meter.
    """
    return multiplier * pow(temp, exponent)

def calculate_dynamic_pressure(air_density, speed):
    """
    Calculate the dynamic pressure imparted on a solid moving through a fluid.

    Args:
    - air_density (float): The density of the fluid.
    - speed (float): The relative speed of the solid and the fluid.

    Returns:
    - float: The dynamic pressure on the solid.
    """
    return 0.5 * air_density * (speed ** 2)

def mach_number_fn(v, temp):
    """
    Calculate the Mach number of an object moving through air at a given temperature.

    Args:
    - v (float): Velocity of the object relative to the air in meters per second.
    - temp (float): Air temperature in Kelvin.

    Returns:
    - float: Mach number (dimensionless).
    """
    return v / np.sqrt(con.adiabatic_index_air_times_R_specific_air * temp)

# gravity
def get_local_gravity(latitude, h = 0, h_expected = None):
    """
    Calculate the acceleration due to gravity at a given latitude and height above sea level.

    Args:
    - latitude (float): Latitude of launch site in degrees.
    - h (float): Ground level elevation above sea level in meters. Defaults to 0.
    - h_expected (float): Expected height the rocket will reach above sea level in meters. Defaults to None, in which case the function will return the acceleration due to gravity at the launchpad instead of the average between the launchpad and the expected apogee.

    Returns:
    - float: Acceleration due to gravity in meters per second squared.

    Based on the International Gravity Formula 1980 (IGF80) model, as outlined in https://en.wikipedia.org/wiki/Theoretical_gravity#International_gravity_formula_1980
    """

    gamma_a = 9.780327  # m/s^2
    c1 = 0.0052790414
    c2 = 0.0000232718
    c3 = 0.0000001262
    c4 = 0.0000000007

    phi = np.deg2rad(latitude)

    gamma_0 = gamma_a * (1 + c1 * np.sin(phi)**2 + c2 * np.sin(phi)**4 + c3 * np.sin(phi)**6 + c4 * np.sin(phi)**8)


    k1 = 3.15704e-07  # 1/m
    k2 = 2.10269e-09  # 1/m
    k3 = 7.37452e-14  # 1/m^2

    g_launchpad = gamma_0 * (1 - (k1 - k2 * np.sin(phi)**2) * h + k3 * h**2)

    if h_expected:
        g_apogee = gamma_0 * (1 - (k1 - k2 * np.sin(phi)**2) * h_expected + k3 * h_expected**2)
        return (g_launchpad + g_apogee) / 2
    else:
        return g_launchpad

# not used in simulator
def pressure_at_height(h, launchpad_temp, launchpad_pressure):
    """
    Calculate the air pressure at a given height above the launchpad.

    Args:
    - h (float): Height above the launchpad in meters.
    - launchpad_temp (float): Temperature at the launchpad in Kelvin.
    - launchpad_pressure (float): Air pressure at the launchpad in Pascals.

    Returns:
    - float: Air pressure at the given height in Pascals.
    """ # note, could use a specific sim's local gravity and lapse rate to be more accurate
    return launchpad_pressure * pow(
        (1 - (h * con.T_lapse_rate / launchpad_temp)),
        (con.F_gravity / (con.R_specific_air * con.T_lapse_rate))
    )

def air_density_fn(pressure, temp):
    """
    Calculate the density of air at a given pressure and temperature.

    Args:
    - pressure (float): Pressure in Pascals.
    - temp (float): Temperature in Kelvin.

    Returns:
    - float: Air density in kilograms per cubic meter.
    """
    return pressure / (con.R_specific_air * temp)

def lookup_dynamic_viscosity(temp):
    """
    Look up the dynamic viscosity of air at a given temperature.

    Args:
    - temp (float): Temperature in Kelvin.

    Returns:
    - float: Dynamic viscosity in kilograms per meter-second.

    Source of lookup table: https://www.me.psu.edu/cimbala/me433/Links/Table_A_9_CC_Properties_of_Air.pdf
    Temperatures converted from source (Celsius to Kelvin).
    """
    # Lookup table for dynamic viscosity
    temps = np.array([173.15, 223.15, 233.15, 243.15, 253.15, 263.15, 273.15, 278.15, 283.15, 288.15, 293.15, 298.15, 303.15, 308.15, 313.15, 318.15, 323.15, 333.15, 343.15])
    viscosities = np.array([1.189e-6, 1.474e-5, 1.527e-5, 1.579e-5, 1.630e-5, 1.680e-5, 1.729e-5, 1.754e-5, 1.778e-5, 1.802e-5, 1.825e-5, 1.849e-5, 1.872e-5, 1.895e-5, 1.918e-5, 1.941e-5, 1.963e-5, 2.008e-5, 2.052e-5])
    
    # Interpolate
    viscosity = np.interp(temp, temps, viscosities)
    return viscosity

def calculate_reynolds_number(air_density, speed, len_characteristic, dynamic_viscosity):
    """
    Calculate the Reynolds number of a solid moving through air.

    Args:
    - air_density (float): The density of the air.
    - speed (float): The relative speed of the solid and the air.
    - len_characteristic (float): The characteristic length of the solid.
    - dynamic_viscosity (float): The dynamic viscosity of the air.

    Returns:
    - float: The Reynolds number of the solid.
    """
    return (air_density * speed * len_characteristic) / dynamic_viscosity

# motor burn curves
    # both likely to either not be used in simulator on the MCU (precompute thrusts and masses at each time given a chosen timestep), or to be rewritten to be more efficient
def mass_at_time(time, dry_mass, fuel_mass_lookup):
    """
    Calculate the total mass of the rocket at a given time during motor burn.

    Args:
    - time (float): Time in seconds since motor ignition.
    - dry_mass (float): Dry mass of the rocket in kilograms.
    - fuel_mass_lookup (dict): Dictionary mapping times to fuel masses.

    Returns:
    - float: Total mass of the rocket at the specified time.
    """
    time_list = list(fuel_mass_lookup.keys())
    lower_time = max([t for t in time_list if t <= time])
    upper_time = min([t for t in time_list if t > time])
    lower_mass = fuel_mass_lookup[lower_time]
    upper_mass = fuel_mass_lookup[upper_time]
    return (dry_mass + lower_mass + (time - lower_time) * (upper_mass - lower_mass) / (upper_time - lower_time))
def thrust_at_time(time, engine_thrust_lookup): 
    """
    Calculate the thrust of the rocket engine at a given time during motor burn.

    Args:
    - time (float): Time in seconds since motor ignition.
    - engine_thrust_lookup (dict): Dictionary mapping times to thrust values.

    Returns:
    - float: Thrust of the engine at the specified time.
    """
    time_list = list(engine_thrust_lookup.keys())
    lower_time = max([t for t in time_list if t <= time])
    upper_time = min([t for t in time_list if t > time])
    lower_thrust = engine_thrust_lookup[lower_time]
    upper_thrust = engine_thrust_lookup[upper_time]
    return lower_thrust + (time - lower_time) * (upper_thrust - lower_thrust) / (upper_time - lower_time)



from .constants import *
from .helper_functions import *
from .rocket_classes import Motor, Rocket, LaunchConditions, Airbrakes
from .flight_simulation import simulate_flight, simulate_airbrakes_flight
from .plotting_functions import plot_ascent, plot_aerodynamics, plot_airbrakes_ascent, display_apogee_parameters_table, create_flight_event_table



# Define Motor, Rocket, LaunchConditions, and Airbrakes classes

import numpy as np

from . import constants as con
from . import helper_functions as hfunc

class Motor:
    """
    The Motor class is used to store the properties of a rocket motor. The properties are:

    - dry_mass: mass of the motor without fuel (kg)
    - thrust_curve: dictionary of thrust (N) at time (s after ignition)
    - total_impulse: total impulse of the motor (Ns)
    - burn_time: time it takes for the motor to burn all of its fuel (s)
    - fuel_mass_curve: dictionary of mass (kg) at time (s after ignition)
    - fuel_mass: total mass of fuel in the motor before ignition (kg)

    If fuel_mass_curve is not provided but feul_mass is, fuel_mass_curve is calculated from the thrust_curve and fuel_mass (assuming fuel burn is proportional to thrust). If fuel_mass_curve is provided, fuel_mass is set to the initial mass in fuel_mass_curve. If neither are provided, fuel_mass and fuel_mass_curve are set to 0.
    """
    
    def __init__(self, dry_mass, thrust_curve, fuel_mass_curve=None, fuel_mass=None):
        self.dry_mass = dry_mass
        self.thrust_curve = thrust_curve

        self.total_impulse = np.trapz(list(thrust_curve.values()), list(thrust_curve.keys()))
        self.burn_time = max(thrust_curve.keys())
        
        if fuel_mass_curve:
            self.fuel_mass_curve = fuel_mass_curve
            self.fuel_mass = fuel_mass_curve[0]
        elif fuel_mass:
            self.fuel_mass = fuel_mass
            self.fuel_mass_curve = {0: fuel_mass}
            times = list(self.thrust_curve.keys())
            for t in range(1, len(times)):
                self.fuel_mass_curve[times[t]] = self.fuel_mass_curve[times[t-1]] - (thrust_curve[times[t]] + thrust_curve[times[t-1]])/2 * (times[t] - times[t-1]) / self.total_impulse * fuel_mass
        else:
            self.fuel_mass = 0
            self.fuel_mass_curve = {
                0: 0,
                self.burn_time: 0
                }


class Rocket:
    """
    The Rocket class is used to store the properties of a rocket.

    Attributes
    ----------
    rocket_mass : float
        Dry mass of the rocket without the motor (kg).
    motor : Motor object
        The rocket's motor.
    A_rocket : float
        Cross-sectional area of the rocket (m^2) used when Cd_rocket_at_Ma was calculated.
    Cd_rocket_at_Ma : float or function
        Coefficient of drag of the rocket. May be given as a function of Mach number or as a constant.
    h_second_rail_button : float
        Height of the second rail button from the bottom of the rocket (m). This is the upper button if there are only 2.
    dry_mass : float
        Total mass of the rocket without fuel (kg).
    Cd_A_rocket : function
        Coefficient of drag of the rocket multiplied by the cross-sectional area of the rocket (m^2).
    """

    def __init__(
        self,
        rocket_mass : float,
        motor : Motor,
        A_rocket : float,
        Cd_rocket_at_Ma = 0.45,
        h_second_rail_button : float = 0.69,
    ):
        """Initializes the Rocket object.

        Parameters
        ----------
        rocket_mass : float
            Dry mass of the rocket without the motor (kg).
        motor : Motor object
            The rocket's motor.
        A_rocket : float
            Cross-sectional area of the rocket (m^2). Must be the same used when the Cd_rocket_at_Ma was calculated.
        Cd_rocket_at_Ma : float or function, optional
            Coefficient of drag of the rocket. May be given as a function of Mach number or as a constant. Defaults to a constant 0.45, which is in the ballpark of what most student team competition rockets our size have.
        h_second_rail_button : float, optional
            Height of the second rail button from the bottom of the rocket (m). This is the upper button if there are only 2. Defaults to 0.7m, which is reasonable for most student team competition rockets. Doesn't matter much if it's not set as it changes apogee by less than 10ft on a 10k ft launch when set to 0.
        """

        self.rocket_mass = rocket_mass
        self.motor = motor
        self.A_rocket = A_rocket
        self.Cd_rocket_at_Ma = Cd_rocket_at_Ma
        self.h_second_rail_button = h_second_rail_button

        self.dry_mass = rocket_mass + motor.dry_mass
        
        if callable(Cd_rocket_at_Ma):
            def Cd_A_rocket_fn(Ma):
                return Cd_rocket_at_Ma(Ma) * A_rocket
        else:
            def Cd_A_rocket_fn(Ma): return Cd_rocket_at_Ma * A_rocket
            # TODO: make it actually operate as a constant if it's not a function
        self.Cd_A_rocket = Cd_A_rocket_fn


class LaunchConditions:
    """The LaunchConditions class is used to store the properties of the launch conditions. 

    Attributes
    ----------
    launchpad_pressure : float
        Pressure at the launchpad (Pa).
    launchpad_temp : float
        Temperature at the launchpad (°C).
    L_launch_rail : float
        Length of the launch rail (m).
    launch_angle : float
        Launch angle from horizontal (deg).
    local_gravity : float
        Acceleration due to gravity at the launch site (m/s^2).
    local_T_lapse_rate : float
        Temperature lapse rate at the launch site (°C/m, K/m).
    """

    def __init__(
        self, 
        launchpad_pressure: float,
        launchpad_temp: float,
        L_launch_rail: float,
        launch_angle: float,
        local_gravity: float = None,
        local_T_lapse_rate: float = con.T_lapse_rate,
        latitude: float = None,
        altitude: float = 0,
    ):
        """Initializes the LaunchConditions object. 
        
        Parameters
        ----------
        launchpad_pressure : float
            Pressure at the launchpad (Pa).
        launchpad_temp : float
            Temperature at the launchpad (°C).
        L_launch_rail : float
            Length of the launch rail (m).
        launch_angle : float
            Launch angle from horizontal (deg).
        local_gravity : float, optional
            Acceleration due to gravity at the launch site (m/s^2). Defaults to 9.80665.
        local_T_lapse_rate : float, optional
            Temperature lapse rate at the launch site (°C/m, K/m). Defaults to -0.0065.
        latitude : float, optional
            Latitude of the launch site (deg). Used along with altitude to calculate local gravity if local_gravity is not provided. If neither local_gravity nor latitude are provided, local gravity defaults to 9.80665. Defaults to None.
        altitude : float, optional
            Altitude of the launch site (m ASL). Used along with latitude to calculate local gravity if local_gravity is not provided. Defaults to 0.
        """
        self.launchpad_pressure = launchpad_pressure
        self.launchpad_temp = launchpad_temp + 273.15
        self.L_launch_rail = L_launch_rail
        self.launch_angle = launch_angle

        self.local_T_lapse_rate = local_T_lapse_rate
        
        if local_gravity:
            self.local_gravity = local_gravity
        elif latitude:
            self.local_gravity = hfunc.get_local_gravity(latitude, altitude)
        else:
            self.local_gravity = con.F_gravity
        

class Airbrakes:
    """
    The Airbrakes class is used to store the properties of the airbrakes.

    Attributes
    ----------
    num_flaps : int
        Number of airbrake flaps.
    A_flap : float
        Cross-sectional area of each flap (m^2).
    Cd_brakes : float
        Coefficient of drag of the airbrakes.
    max_deployment_speed : float
        Maximum speed at which the airbrakes can be deployed (deg/s).
    max_deployment_angle : float
        Maximum angle that the flaps can deploy to (deg).
    """

    def __init__(
        self, 
        num_flaps : int,
        A_flap : float,
        Cd_brakes : float,
        max_deployment_speed : float,
        max_deployment_angle : float,
    ):
        """Initializes the Airbrakes object.

        Parameters
        ----------
        num_flaps : int
            Number of airbrake flaps.
        A_flap : float
            Cross-sectional area of each flap (m^2).
        Cd_brakes : float
            Coefficient of drag of the airbrakes.
        max_deployment_speed : float
            Maximum speed at which the airbrakes can be deployed (deg/s).
        max_deployment_angle : float
            Maximum angle that the flaps can deploy to (deg).
        """
        self.num_flaps = num_flaps
        self.A_flap = A_flap
        self.Cd_brakes = Cd_brakes
        self.max_deployment_speed = max_deployment_speed
        self.max_deployment_angle = max_deployment_angle


class PastFlight ():
    """
    Stores the rocket, launch conditions, and apogee of a past flight
    likely add more things like max speed, max acceleration later
    """
    
    def __init__(self, rocket, launch_conditions, apogee, name = None):
        self.rocket = rocket
        self.launch_conditions = launch_conditions
        self.apogee = apogee
        self.name = name


from setuptools import setup, find_packages

setup(
    name='rocketflightsim',
    version='0.1.1',
    description='A lightweight rocket flight simulator.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Giorgio Chassikos, Western Engineering Rocketry Team',
    author_email='werocketry@gmail.com',
    url='https://github.com/werocketry/RocketFlightSim',
    packages=find_packages(),
    install_requires=[
        "numpy",
        "pandas",
        "matplotlib"
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    license='MIT'
)


import sys
import os

import numpy as np

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from rocketflightsim.rocket_classes import Motor, Rocket, LaunchConditions, Airbrakes

# Example motor configurations
Cesaroni_7579M1520_P = Motor(
    # source: https://www.thrustcurve.org/simfiles/5f4294d20002e900000006b1/
    dry_mass = 2.981,
    thrust_curve = {
        0: 0,
        0.04: 1427.8,
        0.082: 1706.39,
        0.176: 1620.49,
        0.748: 1734.25,
        1.652: 1827.11,
        2.676: 1715.68,
        3.89: 1423.15,
        4.399: 1404.58,
        4.616: 661.661,
        4.877: 69.649,
        4.897: 0
    },
    fuel_mass_curve = {
        0: 3.737,
        0.04: 3.72292,
        0.082: 3.69047,
        0.176: 3.61337,
        0.748: 3.14029,
        1.652: 2.34658,
        2.676: 1.45221,
        3.89: 0.512779,
        4.399: 0.157939,
        4.616: 0.0473998,
        4.877: 0.000343417,
        4.897: 0
    }
)

Cesaroni_7450M2505_P = Motor(
    dry_mass = 2.866, # source: http://www.pro38.com/products/pro98/motor/MotorData.php?prodid=7450M2505-P
    thrust_curve = { # source: https://www.thrustcurve.org/simfiles/5f4294d20002e900000005a0/
        0: 0,
        0.12: 2600,
        0.21: 2482,
        0.6: 2715,
        0.9: 2876,
        1.2: 2938,
        1.5: 2889,
        1.8: 2785,
        2.1: 2573,
        2.4: 2349,
        2.7: 2182,
        2.99: 85,
        3: 0
    },
    fuel_mass_curve = { # source: https://www.thrustcurve.org/simfiles/5f4294d20002e900000005a0/
        0: 3.423,
        0.12: 3.35069,
        0.21: 3.24469,
        0.6: 2.77495,
        0.9: 2.38622,
        1.2: 1.98198,
        1.5: 1.57684,
        1.8: 1.18234,
        2.1: 0.809811,
        2.4: 0.467594,
        2.7: 0.152563,
        2.99: 0.000196996,
        3: 0
    }
)

# Example rocket configuration
example_rocket = Rocket(
    A_rocket = 0.015326, # m^2
    rocket_mass = 14, # kg
    motor = Cesaroni_7579M1520_P,
    Cd_rocket_at_Ma = 0.4,
    h_second_rail_button = 0.69  # m
)

# LaunchConditions class configuration for Spaceport America Cup
T_lapse_rate_SA = -0.00817  # K/m
""" How T_lapse_rate at Spaceport America was determined

Only one source was found with the lapse rate for Spaceport America:
    - https://egusphere.copernicus.org/preprints/2023/egusphere-2023-633/egusphere-2023-633.pdf
    - luckily, they took their measurements in June
    - The lapse rates for the stratosphere for each of three flights were reported as follows:
        - June 1st 2021 -8.4 K/km
        - June 4th 2021 -7.9 K/km
        - June 6th 2021 -8.2 K/km
    - An average of these is what was chosen for the simulation
    - The linear lapse rate was valid for the first 10 km AGL

For future reference, it should be noted that time of year has a large effect on the lapse rate, as reported in:
    - https://mdpi-res.com/d_attachment/remotesensing/remotesensing-14-00162/article_deploy/remotesensing-14-00162.pdf?version=1640917080
    - https://hwbdocs.env.nm.gov/Los%20Alamos%20National%20Labs/TA%2004/2733.PDF
        - states that the average lapse rate in NM is:
            - -4.0F/1000ft (-7.3 K/km) in July
            - -2.5F/1000ft (-4.6 K/km) in January
        - -8.2 K/km is higher than the summer average, but generally desert areas have higher-than-normal lapse rates

The following was the most comprehensive source found for temperature lapse rates in New Mexico: 
- https://pubs.usgs.gov/bul/1964/report.pdf
- No values were found for Spaceport itself, but values for other locations in New Mexico were found
- the report says that in the western conterminous United States, temperature lapse rates are generally significantly less than the standard -6.5 K/km
- the report didn't include the date (or month) of the measurements, so I'd assume that it happened in the winter due to the low lapse rates, and/or the data being several decades old means that it's no longer as accurate due to the changing global climate
- has values for many locations in New Mexico (search for n. mex), and they ranged from -1.4 to -3.9 K/km
    - the closest station to SC was Datil, which had a lapse rate of -3.1 K/km
"""
L_launch_rail_ESRA_provided_SAC = 5.18  # m, 
""" ESRA provides teams with a 5.18m rail at competition """
launchpad_pressure_SAC = 86400  # Pa
""" How the launchpad pressure at Spaceport America was determined

- 86400 2022/06/24   WE Rocketry 2022 TeleMega/TeleMetrum data
- 86405 2022/06/23   https://github.com/ISSUIUC/flight-data/tree/master/20220623
- 86170 2023/06/21   https://github.com/ISSUIUC/flight-data/tree/master/20230621
"""
launchpad_temp_SAC = 35  # deg C
""" Ground-level temperature at Spaceport America Cup note

Flights can occur between about 07:00 and 16:30 local time, so the temperature at the time of launch can vary significantly. 35 C is about what it has been historically during the competition in mid-late June. Getting closer to launch day, it would be more accurate to use a weather forecast to get a value for expected temperature(s).

You can also consider running simulations with a range of temperatures that have been seen on launch days in the past (normally between 25 and 45 C) to see how different ground-level temperatures could affect a rocket's flight.
"""
latitude_SA = 32.99  # deg, Spaceport America's latitude
""" https://maps.app.goo.gl/rZT6MRLqHneA7wNX7 """
altitude_SA = 1401  # m, Spaceport America's elevation
""" https://www.spaceportamerica.com/faq/#toggle-id-15 """
launch_angle_SAC = 84  # deg from horizontal
""" How the standard launch angle at Spaceport America Cup was determined
DTEG 10.1.1: 
    > Launch vehicles will nominally launch at an elevation angle of 84° ±1°
DTEG 10.1.2:
    > Range Safety Officers reserve the right to require certain vehicles’ launch elevation be 
lower or higher if flight safety issues are identified during pre-launch activities

Teams have noted that they've been told to use angles at least as low as 80°. The Range Safety Officer picks the angle based on various factors, including the rocket being launched, the weather, and the location of the launch pad. In the design, simulation, and testing phases, use the nominal angle of 84°, but consider the possibility of the launch angle being more or less than that on competition day.
"""

Spaceport_America_avg_launch_conditions = LaunchConditions(
    launchpad_pressure = launchpad_pressure_SAC,
    launchpad_temp = launchpad_temp_SAC,
    L_launch_rail = L_launch_rail_ESRA_provided_SAC,
    launch_angle = launch_angle_SAC,
    local_T_lapse_rate = T_lapse_rate_SA,
    latitude = latitude_SA,
    altitude = altitude_SA
)

# Default airbrakes model
default_airbrakes_model = Airbrakes(
    num_flaps = 3,
    A_flap = 0.004,  # m^2  flap area
    Cd_brakes = 1,
    max_deployment_speed = 5,  # deg/s
    max_deployment_angle = 45  # deg
)

if __name__ == "__main__":
    from rocketflightsim.flight_simulation import simulate_flight, simulate_airbrakes_flight
    from rocketflightsim.plotting_functions import plot_ascent, plot_aerodynamics, plot_airbrakes_ascent

    dataset, liftoff_index, launch_rail_cleared_index, burnout_index, apogee_index = simulate_flight(rocket = example_rocket, launch_conditions = Spaceport_America_avg_launch_conditions, timestep = 0.001)
    print(f"Motor burnout: \n\tHeight: {dataset['height'].iloc[burnout_index - 1]} m\n\tSpeed: {dataset['speed'].iloc[burnout_index - 1]} m/s\n\tTime: {dataset['time'].iloc[burnout_index - 1]} s\n")
    print(f"Apogee: \n\tHeight: {dataset['height'].iloc[apogee_index - 1]} m\n\tTime: {dataset['time'].iloc[apogee_index - 1]} s\n")

    plot_ascent(dataset['time'], dataset['height'], dataset['speed'], dataset['v_y'], dataset['a_y'])


# constants 
R_universal = 8.3144598  # J/(mol*K)
MM_air = 0.0289644  # kg/mol
adiabatic_index_air = 1.4  # unitless
"""Notes on adiabatic index (also known as the heat capacity ratio (cp/cv)) for air

As a function of teperature, for dry air, the adiabatic index according to different sources is:
    https://en.wikipedia.org/wiki/Heat_capacity_ratio
    
        - 1.404 at -15°C
        - 1.403 at 0°C
        - 1.400 at 20°C
        - 1.398 at 200°C

    https://www.chemeurope.com/en/encyclopedia/Heat_capacity_ratio.html

        - 1.403 at 0°C
        - 1.400 at 20°C
        - 1.401 at 100°C
        
The value of 1.4 is a very good approximation for the temperature range the rocket will experience

As per https://www.grc.nasa.gov/WWW/BGH/realspec.html, air is calorically perfect up to low supersonic Mach numbers, so the adiabatic index doesn't change with airspeed over the speeds that the rocket will experience
"""

# derived constants
R_specific_air = R_universal / MM_air  # J/(kg*K)
adiabatic_index_air_times_R_specific_air = adiabatic_index_air * R_specific_air  # J/(kg*K)

# conversion factors
m_to_ft_conversion = 3.28084  # ft/m

# default launch site values. Specific launch site values can be set in LaunchConditions objects
F_gravity = 9.80665  # m/s^2
T_lapse_rate = -0.0065  # K/m